#-------------------------------------------------------------------------------
#
#	Class: WavFile
#
class WavFile

	attr_reader :name
	attr_reader :freq
	attr_reader :sbit
	attr_reader :maxSmp

	def initialize(name)
		@file	= open(@name = name)
		@riff	= @file.read(4)
		@tsize	= @file.read(4).unpack('V')[0]
		@wave	= @file.read(4)
		@fmt_	= @file.read(4)
		@hsize	= @file.read(4).unpack('V')[0]
		@type	= @file.read(2).unpack('v')[0]
		@ch		= @file.read(2).unpack('v')[0]
		@freq	= @file.read(4).unpack('V')[0]
		@rate	= @file.read(4).unpack('V')[0]
		@block	= @file.read(2).unpack('v')[0]
		@sbit	= @file.read(2).unpack('v')[0]
		@dsize	= 0; begin
			@dummy	= @file.read(@dsize)
			@data	= @file.read(4)
			@dsize	= @file.read(4).unpack('V')[0]
		end until(@data.downcase == 'data')
		@maxBlk	= @dsize / @block
		@sbyte	= @sbit >> 3
		@smp	= @sbyte * @ch
		@maxSmp	= @dsize / @smp
		@top	= @file.pos
	end

	#---------------------------------------------------------------------------
	#
	#	Get wavfile infomation
	#
	def get_info
		info = []
		info.push "FileName : #{@name}"
		info.push "FileKind : #{@riff}"
		info.push "FileSize : #{@tsize}"
		info.push "FileType : #{@wave}"
		info.push "FileForm : #{@fmt_}"
		info.push "HeadSize : #{@hsize}"
		info.push "HeadType : #{@type}"
		info.push "Channel  : #{@ch}"
		info.push "Frequency: #{@freq}"
		info.push "DataRate : #{@rate}"
		info.push "BlockSize: #{@block}"
		info.push "SampleBit: #{@sbit}"
		info.push "DataType : #{@data}"
		info.push "DataSize : #{@dsize}"
		info.push "TotalTime: #{Span.to_s(Float(@maxSmp) / @freq)}"
		maxLen = 0
		info.each {|i|
			maxLen = i.size if maxLen < i.size					# 表示に必要な列数(横幅)を得る
		}
		[info, info.size, maxLen]								# 情報の配列、表示必要行、表示必要列
	end

	#---------------------------------------------------------------------------
	#
	#	Save top
	#
	def save_top(wFile, blocks)
		wFile.write  @riff
		wFile.write [blocks * @block + 36].pack('V')
		wFile.write  @wave
		wFile.write  @fmt_
		wFile.write [@hsize].pack('V')
		wFile.write [@type ].pack('v')
		wFile.write [@ch   ].pack('v')
		wFile.write [@freq ].pack('V')
		wFile.write [@rate ].pack('V')
		wFile.write [@block].pack('v')
		wFile.write [@sbit ].pack('v')
		wFile.write  @data
		wFile.write [remain = blocks * @block].pack('V')
		remain
	end

	#---------------------------------------------------------------------------
	#
	#	Save phrase
	#
	def save_phrase(startBlock, blocks, wFileName)
		wFile = open(wFileName, 'w')
		remain = saveBytes = save_top(wFile, blocks)

		@file.pos = @top + startBlock * @block
		loop {
			chank = remain > 65536 ? 65536 : remain
			wFile.write(@file.read(chank)); remain -= chank		# 小分けに書き出す
			yield((saveBytes - remain) * 100 / saveBytes)		# 進捗率(%)を返す
			break if remain == 0
		}
		wFile.close
	end

	#---------------------------------------------------------------------------
	#
	#	Save phrase reverse
	#
	def save_phrase_reverse(startBlock, blocks, wFileName)
		wFile = open(wFileName, 'w')
		remain = saveBytes = save_top(wFile, blocks)

		pmode = @ch == 2 ? 'i*' : 's*'; top2 = @top + startBlock * @block
		loop {
			chank = remain > 65536 ? 65536 : remain
			@file.pos = top2 + remain - chank
			rev = @file.read(chank).unpack(pmode).reverse.pack(pmode)	# 逆転して
			wFile.write(rev); remain -= chank					# 小分けに書き出す
			yield((saveBytes - remain) * 100 / saveBytes)		# 進捗率(%)を返す
			break if remain == 0
		}
		wFile.close
	end

	#---------------------------------------------------------------------------
	#
	#	Save phrase with filter
	#
	def save_phrase_filter(startBlock = 0, blocks = @maxBlk, wFileName = nil)
		wFile = (it = wFileName) ? it : $stdout
		wFile = open(wFileName, 'w') if(wFileName.is_a?(String))
		remain = saveBytes = save_top(wFile, blocks)

		last_prog = nil; blocks.times {|p|
			it = p * 100 / blocks; prog = (it != last_prog) ? it : nil; last_prog = it
			gain = yield(get_gain(startBlock + p) << prog)
			wFile.write(encode_gain(gain))
		}
		wFile.close if(wFileName.is_a?(String))
	end

	#---------------------------------------------------------------------------
	#
	#	Get sample
	#
	def get_sample(markPoint, step, smps, offset)
		sample = []; smps.times {|x|
			samplePoint = markPoint + (x + offset) * step
			sample << get_gain(samplePoint).unshift(x)			# 座標情報を付加して記憶
		}
		sample
	end

	#---------------------------------------------------------------------------
	#
	#	Get gain
	#
	def get_gain(p)
		chs = []
		if(-1 < p and p < @maxSmp)
			@file.pos = @top + @smp * p
			@ch.times {
				s = @file.read(@sbyte)
				if(@sbit == 16)
					l = s.unpack('v')[0]
					l > 0x7fff and l -= 0x10000
				elsif(@sbit == 24)
					l = (s + "\x00").unpack('V')[0]
					l > 0x7fffff and l -= 0x1000000
				else
					raise('unexpected sampling bits.')
				end
				chs << l
			}
			chs.size == 1 and chs << chs[0]
			chs.unshift(chs.absmax)								# mixing
		else
			[nil, nil, nil]
		end
	end

	#---------------------------------------------------------------------------
	#
	#	Encode gain
	#
	def encode_gain(gain)
		blk = ''
		@ch.times {|ch|
			if(@sbit == 16)
				(l = gain[ch + 1]) < 0 and l += 0x10000
				b = [l].pack('v')
			elsif(@sbit == 24)
				(l = gain[ch + 1]) < 0 and l += 0x1000000
				b = [l].pack('V')[0, 3]
			else
				raise('unexpected sampling bits.')
			end
			blk << b
		}
		blk
	end

	#---------------------------------------------------------------------------
	#
	#	Get gain range
	#
	def get_gain_range(start, finish, step, deltaY)
		range = {}
		finish = @maxBlk unless finish
		(1..(((65535 * 256) / deltaY) + 1)).each {|i|			# make statistics base
			range[i] = 1
		}
		samplePoint = start; i = 0
		loop {
			break if samplePoint >= finish
			(c, l, r) = get_gain(samplePoint)
			range[(((32768 - c) * 256) / deltaY) + 1] += 1
			samplePoint += step
			yield(samplePoint * 100 / @maxBlk) if ((i += 1) % 240) == 1		# 進捗率(%)を返す
		}
		range
	end

	#---------------------------------------------------------------------------
	#
	#	Auto split
	#
	def auto_split(silent = 1024, step = 11025, repeat = 6, offset = -3)
		marks = []
		samplePoint = 0; rep = 0; i = 0
		loop {
			break if samplePoint >= @maxBlk
			(c, l, r) = get_gain(samplePoint)
			if c.abs < silent
				rep += 1
			else
				marks.push samplePoint + (offset * step) if rep > repeat
				rep = 0
			end
			samplePoint += step
			yield(samplePoint * 100 / @maxBlk) if ((i += 1) % 240) == 1		# 進捗率(%)を返す
		}
		marks
	end

	#---------------------------------------------------------------------------
	#
	#	Get sample
	#
	def xget_sample(markPoint, step, smps, offset)
		sample = []
		(0..smps - 1).each {|x|
			samplePoint = markPoint + (x + offset) * step
			sample.push get_gain(samplePoint).unshift(x)		# 座標情報を付加して記憶
		}
		sample
	end
end

#-------------------------------------------------------------------------------
#
#	Class: NewWavFile
#
class NewWavFile < WavFile

	attr_reader :samples

	def initialize(name)
		@samples = []
		super
	end

	def []=(b, sample)
		@samples[b] = [sample[0], sample[1] || sample[0]]
	end

	def <<(sample)
		@samples << sample
	end

	#---------------------------------------------------------------------------
	#
	#	Save phrase
	#
	def save_phrase(startBlock, blocks, wFileName)
		blocks ||= @samples.size
		wFile = open(wFileName, 'w')
		wFile.write  @riff
		wFile.write [blocks * @block + 36].pack('V')
		wFile.write  @wave
		wFile.write  @fmt_
		wFile.write [@hsize].pack('V')
		wFile.write [@type ].pack('v')
		wFile.write [@ch   ].pack('v')
		wFile.write [@freq ].pack('V')
		wFile.write [@rate ].pack('V')
		wFile.write [@block].pack('v')
		wFile.write [@sbit ].pack('v')
		wFile.write  @data
		wFile.write [remain = saveBytes = blocks * @block].pack('V')

		(startBlock...(startBlock + blocks)).each {|b|
			wFile.write(@samples[b].pack('s*'))
		}
		wFile.close
	end
end

#-------------------------------------------------------------------------------
#
#	Class: Array - absmax
#
class Array
	def absmax
		max = self[0]
		self.each {|x|
			max = x if max.abs < x.abs
		}
		max
	end
end

#-------------------------------------------------------------------------------
#
#	Class: Span
#
class Span
	def self.to_s(n)
		sprintf('%3d:%05.2f', n / 60, (n % 60).abs)				# % bug ?
	end
end

__END__

